﻿#region Using ステートメント

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;

#endregion

namespace WpfFontSample
{
    /// <summary>
    /// WPFフォントプロセッサー、サンプルプログラム
    /// </summary>
    public class WpfFontSampleGame : Microsoft.Xna.Framework.Game
    {
        #region フィールド

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // スプライトフォント
        SpriteFont font;
        SpriteFont titleFont;
        SpriteFont gradientFont;
        SpriteFont outlineFont;

        // 文章をスクロール表示する為のフィールド
        float scrollOffset = 0;         // スクロール位置
        float scrollSpeed = 0;          // スクロール速度(Windows Phoneのフリックで使用)
        float maxScrollOffset = 0;      // スクロール上限値

        // 入力処理用のマウスステート
        MouseState prevMouseState;

        #endregion

        #region 初期化

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public WpfFontSampleGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

#if WINDOWS_PHONE
            graphics.IsFullScreen = true;
            graphics.PreferredBackBufferWidth = 800;
            graphics.PreferredBackBufferHeight = 480;

            // Windows Phoneで60FPSで動作させるための設定
            TargetElapsedTime = TimeSpan.FromTicks(166666);
            graphics.PreparingDeviceSettings +=
                new EventHandler<PreparingDeviceSettingsEventArgs>(
                    graphics_PreparingDeviceSettings);
#else
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;
#endif

            // タッチジェスチャーの登録
            TouchPanel.EnabledGestures =
                GestureType.Flick | GestureType.VerticalDrag | GestureType.Tap;

        }

        // Windows Phoneで60FPSで動作させるための設定
        void graphics_PreparingDeviceSettings(object sender,
            PreparingDeviceSettingsEventArgs e)
        {
            e.GraphicsDeviceInformation.PresentationParameters.PresentationInterval
                = PresentInterval.One;
        }

        /// <summary>
        /// コンテントの読み込み
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // フォントデータの読み込み
            font = Content.Load<SpriteFont>("font");
            titleFont = Content.Load<SpriteFont>("titleFont");
            gradientFont = Content.Load<SpriteFont>("gradientFont");
            outlineFont = Content.Load<SpriteFont>("outlineFont");

            // 段落データの調整とスクロール上限値の計算
            maxScrollOffset = 0;
            foreach (var paragraph in MessageTable.Messages)
            {
                // フォント設定
                switch(paragraph.FontType)
                {
                    case MessageFontType.Normal:    paragraph.Font = font; break;
                    case MessageFontType.Title:     paragraph.Font = titleFont; break;
                    case MessageFontType.Gradient:  paragraph.Font = gradientFont; break;
                    case MessageFontType.Outline:   paragraph.Font = outlineFont; break;
                }

                // 段落サイズの計算
                paragraph.Size = paragraph.Font.MeasureString(paragraph.Text);

                // スクロール位置調整
                maxScrollOffset -= paragraph.Size.Y;
            }

            // 最後のスクロール位置で画面半分まで文字が見えるように調整する
            maxScrollOffset += GraphicsDevice.Viewport.Bounds.Height * 0.5f;
            maxScrollOffset = (int)Math.Ceiling(maxScrollOffset);
        }

        #endregion

        #region 更新と描画

        /// <summary>
        /// ゲームの更新(衝突判定、入力処理、オーディオの再生など)
        /// </summary>
        /// <param name="gameTime">タイミング値のスナップショット</param>
        protected override void Update(GameTime gameTime)
        {
            float dt = (float)gameTime.ElapsedGameTime.TotalSeconds;

            // ゲーム終了処理
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // 入力処理

            // ゲームパッド処理
            for (int i = 0; i < 4; ++i)
            {
                var padState = GamePad.GetState((PlayerIndex)i);

                scrollOffset += padState.ThumbSticks.Right.Y * 20.0f;
                scrollOffset += padState.ThumbSticks.Left.Y * 20.0f;
            }

            // キーボード入力
            var keyboardState = Keyboard.GetState();
            if (keyboardState.IsKeyDown(Keys.Down))
                scrollOffset -= 15;
            if (keyboardState.IsKeyDown(Keys.Up))
                scrollOffset += 15;

            // マウスのホイールスクロール処理
            var mouseState = Mouse.GetState();
            scrollOffset +=
                (mouseState.ScrollWheelValue - prevMouseState.ScrollWheelValue) * 0.3f;
            prevMouseState = mouseState;

            // タッチ入力処理
            while (TouchPanel.IsGestureAvailable)
            {
                var gesture = TouchPanel.ReadGesture();

                switch(gesture.GestureType)
                {
                    case GestureType.Flick:
                        scrollSpeed = gesture.Delta.Y;
                        break;
                    case GestureType.VerticalDrag:
                        scrollSpeed = 0;
                        scrollOffset += gesture.Delta.Y;
                        break;
                    case GestureType.Tap:
                        scrollSpeed = 0;
                        break;
                }
            }

            // 文字スクロール速度を減速させる
            scrollSpeed = MathHelper.Lerp(scrollSpeed, 0, dt);
            if (Math.Abs(scrollSpeed) < 20)
                scrollSpeed = 0;

            // スクロール位置更新
            scrollOffset += scrollSpeed * dt;
            scrollOffset = MathHelper.Clamp(scrollOffset, maxScrollOffset, 0);

            base.Update(gameTime);
        }

        /// <summary>
        /// ゲームの描画
        /// </summary>
        /// <param name="gameTime">タイミング値のスナップショット</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // 各段落の描画

#if WINDOWS_PHONE
            Vector2 pos = new Vector2(20, 20);
#else
            Vector2 pos = new Vector2(240, GraphicsDevice.Viewport.TitleSafeArea.Top);
#endif
            // スクロール位置の設定
            pos.Y += (int)scrollOffset;

            // 画面領域の取得
            var viewBounds = GraphicsDevice.Viewport.Bounds;
            var paragraphBounds = new Rectangle{ X = (int)pos.X };

            spriteBatch.Begin();
            for (int i = 0; i < MessageTable.Messages.Length; ++i)
            {
                var paragraph = MessageTable.Messages[i];

                // 現在の段落が画面内にある場合にのみ描画する
                paragraphBounds.Y = (int)pos.Y;
                paragraphBounds.Width = (int)paragraph.Size.X;
                paragraphBounds.Height = (int)paragraph.Size.Y;

                if (viewBounds.Intersects(paragraphBounds))
                {
                    spriteBatch.DrawString(paragraph.Font, paragraph.Text,
                                            pos, Color.White);
                }

                // 表示位置を段落の高さ分だけ進める
                pos.Y += paragraph.Size.Y;
            }
            spriteBatch.End();

            base.Draw(gameTime);
        }

        #endregion

    }
}
